; ===============================================================
; AI
; ===============================================================

; ---------------------------------------------------------------
; AITree
; AI decision tree
; Input: (Style)
; Output: B = Move
; ---------------------------------------------------------------

AITree:
	di
	call AITreeRoot		; HL = pt to root of style subtree
	call AITreeNode		; traverse tree, return B = move or nc
	ei
	ret c				; ret if done
	
AITreeRand:
	ld B, 4			; else pick random move
	call iRandom
	ld HL, OppPKMN + PKMN_Moves	; HL = pt to move array
	ld B, 0
	ld C, A
	add HL, BC
	ld B, (HL)
	scf
	ret				; ret w/ c flag
	
AITreeRoot:				; subroutine - gets root
	ld A, (Style)
	ld HL, aiTreeDb
	RCall( cGetTable2 )		; HL = pt to AI tree node (RAM)
	ret
	
; ---------------------------------------------------------------
; AITreeNode
; Performs action / test, returns if NULL children, else branch
; Input: HL = pt to node
; Output: B = move
;         c = done, nc = not done
; ---------------------------------------------------------------

AITreeNode:
	inc HL
	push HL			; save pt to node, first child
	ld A, (HL)			; A = first child (NULL or > 0)
	or A				; z if A is NULL
	push AF			; save flags
	dec HL
	ld A, (HL)			; A = ai action
	ld B, A			; save A as B
	and aiLvl4			; bitmask to isolate lvl bits (bits 6-7)
	rlca
	rlca				; left rotate A twice, bit 7 -> bit 1, bit 6 -> bit 0
	inc A				; A ++, A = aiLvl
	ld HL, RoundNum		; HL = pt to current round number
	ld C, (HL)			; C = current round number
	cp C				; if aiLvl < round number
	jr c, AITreeNodeLvl		; then execute ai node
	pop HL
	pop HL
	jr AITreeRand			; else fail, choose random move, return w/ c flag
	
AITreeNodeLvl:
	ld A, B			; restore A from B
	and ~aiLvl4			; remove lvl bits (bits 6-7)
	ld HL, AIActionDb
	RCall( cGetTable2 )		; HL = routine address
	
	pop AF
	jr z, AITreeNode0
AITreeNode1:
	ld BC, AITreeNodeRet1	; if not zero, then return to Ret1
	jr AITreeNodeCont
	
AITreeNode0:
	ld BC, AITreeNodeRet0	; else if zero, then return to Ret0
	
AITreeNodeCont:
	push BC			; push return address onto stack
	push HL
	ret				; jump to HL
	
AITreeNodeRet0:			; function returns c (success) or nc (fail)
	pop HL				; pop pt off stack
	ret				; return w/ c or nc
	
AITreeNodeRet1:			; function returns c (first child) or nc (second chiild)
	pop HL				; HL = pt to node, first child
	jr c, AITreeNodeFirst	; if nc, second child
AITreeNodeSecond:
	inc HL				; then inc HL to pt to next child
	call AITreeNodeCall		; try second child
	ret c				; ret if done
	dec HL
	jr AITreeNodeCall		; try first child, ret w/ c or nc ; jp
	
AITreeNodeFirst:
	call AITreeNodeCall		; try first child
	ret c				; ret if done
	inc HL
	;jp AITreeNodeCall		; try second child, ret w/ c or nc
	
AITreeNodeCall:			; subroutine - input: HL = pt to child node, output: B = move, c = done, nc = not done, preserves HL
	push HL
	ld A, (HL)			; A = child number
	
	call AITreeNodeGet		; HL = pt to child node
	call AITreeNode		; recursively call AITreeNode
	pop HL
	ret
	
AITreeNodeGet:			; subroutine - input: A = node number, HL = pt to node
	dec A				; node number is 1 indexed
	ld D, 0
	ld E, A			; DE = zero extended A
	push DE
	pop HL				; HL = DE
	add HL, HL
	add HL, DE			; HL = HL * 3
	ex DE, HL			; DE = HL * 3
	ld HL, aiFree			; HL = pt to root of tree
	add HL, DE			; HL = pt to node
	ret

; ---------------------------------------------------------------
; AIActDamage
; Maximizes expected damage - argmax (damage * acc)
; Input: (PlayerPKMN), (OppPKMN)
; Output: B = Move ID
;         c = success, nc = fail
; Interrupts disabled to speed up routine
; ---------------------------------------------------------------

AIActDamage:
	;RCall( cRAIActDamage )
	;ret
	
	ld HL, OppPKMN		; set CalcDamage parameters
	ld (AtkPKMN), HL
	ld HL, PlayerPKMN
	ld (TargetPKMN), HL
	
	ld B, $FF			; B = current best move = none
	ld C, 4			; C = counter = 4
	ld DE, 0			; DE = current max value
	ld HL, OppPKMN + PKMN_Moves	; HL = pt to current move
	
AIActDamageLp:
	push HL
	push BC
	push DE
	ld A, (HL)
	ld (AtkMove), A		; (AtkMove) = A = current move
	push AF			; save current move
	call AIActDamageCalc		; HL = calculated damage
	ld D, 10
	RCall( cDivHL )		; L = HL / 10
	pop AF
	push AF
	push HL			; save damage from move
	call AIActDamageAcc
	pop HL
	
	ld H, A
	call MultHL			; HL = A * L
	pop AF
	pop DE
	pop BC
	
	call CpHLDE			; regs unchanged
	jr c, AIActDamageLT		; if HL >= DE
	ld B, A			; then set new best move
	ld D, H
	ld E, L			; and set new value
AIActDamageLT:			; else if HL < DE, then skip
	pop HL
	inc HL				; next move
	dec C				; dec counter
	jr nz, AIActDamageLp		; loop until C = 0
	
	ld A, D
	or E				; if DE = 0
	ret z				; ret w/ nc
	scf
	ret				; else ret w/ c
	
AIActDamageCalc:			; subroutine - calculates damage, input: A = current move, output: HL = damage
	ld HL, 100
	cp mExplosion			; if move = explosion
	ret z				; then ret w/ HL = 100
	RCall( cCalcDamage )		; HL = damage
	call Attack_DamClamp		; HL = min(damage, target HP)
	xor A
	ld (FocusBandAct), A		; reset focusbandact
	ld (LevitateAct), A		; reset levitateact
	ret
	
AIActDamageAcc:			; subroutine - calculates accuracy, input: A = current move, output: A = accuracy
	push AF			; save A
	RCall( cCalcMiss )		; calculates if hit (c) or miss (nc)
	pop BC				; get A as B
	ld A, 0			; A = 0
	ret nc				; ret w/ A = 0, if calculated as miss
	ld A, B			; otherwise, restore A
	
	RCall( cGetMoveAcc )		; A = move accuracy
	or A				; if A != 0
	ret nz				; then ret
	ld A, 100			; else set A = 100
	ret
	
; ---------------------------------------------------------------
; AIActHeal
; Picks healing move
; Input: (OppPKMN)
; Output: B = move
;         c = success, nc = fail (no move)
; ---------------------------------------------------------------

AIActHeal:
	ld HL, ResearchStyleDbEndure + 1	; skip size byte
	ld C, 7			; 7 elements in database
	
AIActMoveLp:				; subroutine - checks moves against database, ret c if match, nc otherwise. input: C = database size, HL = pt
	ld A, (HL)			; A = move ID
	ex DE, HL			; save pt to database
	ld HL, OppPKMN + PKMN_Moves	; HL = pt to move array
	ld B, 4			; 4 moves in array
	
AIActMoveLp2:
	cp (HL)			; if A is in move array
	jr z, AIActMoveSuccess	; then return
	inc HL				; else get next move in array
	djnz AIActMoveLp2		; loop until B = 0
	
	ex DE, HL			; restore pt to database
	inc HL				; get next move
	dec C				; dec counter
	jr nz, AIActMoveLp		; loop until C = 0
	
	or A				; if no matching moves
	ret				; then ret w/ nc
	
AIActMoveSuccess:
	ld B, A			; return B = move
	scf				; set c for success
	ret
	
; ---------------------------------------------------------------
; AIActBoost
; Returns stat boosting move, or nc 
; ---------------------------------------------------------------

AIActBoost:
	ld HL, ResearchStyleDbPrep + 1
	ld C, 12			; skip over psych up
	jr AIActMoveLp		; try to find matching move
	
; ---------------------------------------------------------------
; AIActStatus
; ---------------------------------------------------------------

AIActStatus:
	ld HL, ResearchStyleDbSlow + 1
	ld C, 11
	jr AIActMoveLp
	
; ---------------------------------------------------------------
; AIActProtect
; ---------------------------------------------------------------

AIActProtect:
	ld HL, ResearchStyleDbEndure + 8
	ld C, 3
	jr AIActMoveLp
	
; ---------------------------------------------------------------
; AIActCure
; ---------------------------------------------------------------

AIActCure:
	ld HL, ResearchStyleDbEndure + 16
	ld C, 2
	jr AIActMoveLp
	
; ---------------------------------------------------------------
; AIActLower
; ---------------------------------------------------------------

AIActLower:
	ld HL, ResearchStyleDbWeaken + 1
	ld C, 8
	jr AIActMoveLp
	
; ---------------------------------------------------------------
; AIActWeather
; ---------------------------------------------------------------

AIActWeather:
	ld HL, ResearchStyleDbAdapt	 + 1
	ld C, 4
	jr AIActMoveLp
	
; ---------------------------------------------------------------
; AIHP50
; ---------------------------------------------------------------

AIChkHP50:
	ld HL, OppPKMN
	jp ChkHalfHP			; input: HL = pt to data, output: c if HP < 1/2 max HP
	
; ---------------------------------------------------------------
; AIChkStatHigh
; Returns c if 1 stat is stage 3 or higher, nc otherwise
; ---------------------------------------------------------------

AIChkStatHigh:
	ld HL, OppPKMN + PKMN_AtkBoost
	ld B, 7
	
AIChkStatHighLp:
	ld A, (HL)			; A = stat boost
	cp 3				; if boost < 3
	jr c, AIChkStatHighNo	; then skip
	cp 7				; if boost < 7
	ret c				; then ret c (3 <= boost <= 6)
	
AIChkStatHighNo:
	inc HL				; HL pts to next stat boost
	djnz AIChkStatHighLp
	or A
	ret				; ret w/ nc if no stat >= 3
	
; ---------------------------------------------------------------
; AIChkTrgtStatus
; ---------------------------------------------------------------

AIChkTrgtStatus:
	ld A, (PlayerPKMN + PKMN_Status)
	
AIChkZero:
	or A				; if A = 0 (stNone)
	ret z				; then ret w/ nc
	scf				; else, ret w/ c
	ret
	
; ---------------------------------------------------------------
; AIChkPrevProtect
; ---------------------------------------------------------------

AIChkPrevProtect:
	ld A, (OppPKMN + PKMN_Counter)
	jr AIChkZero
	
; ---------------------------------------------------------------
; AIChkHasStatus
; ---------------------------------------------------------------

AIChkHasStatus:
	ld A, (OppPKMN + PKMN_Status)
	jr AIChkZero
	
; ---------------------------------------------------------------
; AIChkIsWeather
; ---------------------------------------------------------------

AIChkIsWeather:
	ld A, (Weather)
	jr AIChkZero
	
; ---------------------------------------------------------------
; AIChkStatLow
; ---------------------------------------------------------------

AIChkStatLow:
	ld HL, PlayerPKMN + PKMN_AtkBoost
	ld B, 7
	
AIChkStatLowLp:
	ld A, (HL)			; A = stat boost
	cp 7				; if boost < 7 (0 or positive)
	jr c, AIChkStatLowNo		; then skip
	cp -1				; else if boost < -1
	ret c				; then ret w/ c

AIChkStatLowNo:
	inc HL				; HL pts to next stat boost
	djnz AIChkStatLowLp
	or A
	ret				; ret w/ nc if no stat < -2